
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <sys/types.h>                      /* time_t */
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>                          /* printf() */
#include <assert.h>
#include <errno.h>

#include "initng_history.h"
#include "../../src/initng_global.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../ngc2/initng_ngc2.h"
#include "../../src/initng_static_states.h"

int history_records = 0;

history_h history_db;

static void fetch_output(active_h * service, int datalen,
                         process_h * process);
static void cmd_history(char *arg, FILE * fd);
static void cmd_log(char *arg, FILE * fd);

s_command HISTORYS =
    { 'l', "history", DATA_COMMAND, USES_OPT, {(void *) &cmd_history},
"Print out history_db."
};

s_command LOG = { 'L', "LOG", PRINT_COMMAND, USES_OPT, {(void *) &cmd_log},
"Print out log."
};



static void history_db_compensate_time(time_t skew)
{
    history_h *current = NULL;

    D_("history_db_compensate_time(%i);\n", (int) skew);

    while_history_db(current)
    {
        if (current->time)
            current->time += skew;
    }
}



static void history_db_clear_service(active_h * service)
{
    history_h *current = NULL;

    D_("history_db_clear_service(%s);\n", service->name);

    while_history_db(current)
    {
        if (current->service == service)
        {
            current->service = NULL;
            current->name = i_strdup(service->name);
        }
    }
}

/* clear history */
static void history_free_all(void)
{
    history_h *current, *safe = NULL;

    /* while there is a histry db */
    while_history_db_safe(current, safe)
    {
        /* free histry name */
        if (current->name)
            free(current->name);

        /* free data if any */
        if (current->data)
            free(current->data);

        /* remove from history_db list */
        list_del(&current->list);

        /* free histroy entry */
        free(current);
    }
}

static int add_hist(history_h * hist)
{
    /* add struct */
    list_add(&hist->list, &history_db.list);

    /* check lenght of history_db, and purge it on the end! */
    history_records++;

    if (history_records > HISTORY)
    {

        struct list_head *last = history_db.list.prev;
        history_h *entry = list_entry(last, history_h, list);

        if ((entry) && entry->name)
            free(entry->name);

        list_del(last);
        free(entry);
        history_records--;
    }

    /* leave */
    return (TRUE);
}

/* add values to history database */
static int history_add_values(active_h * service)
{
    history_h *tmp_e = NULL;    /* temporary pointer to insert data and add to db */

    assert(service->name);

    /* Dont bother adding */
    if (!service->a_state)
        return (TRUE);

    /*if (!service->a_state->state_name);
       return(TRUE); */

    D_("adding: %s.\n", service->name);

    /* allocate space for data */
    if (!(tmp_e = (history_h *) initng_calloc(1, sizeof(history_h))))
    {
        F_("Out of memory.\n");
        return (TRUE);
    }

    /* set data in struct */
    tmp_e->service = service;
    tmp_e->name = NULL;
    tmp_e->time = service->time_got_status;
    tmp_e->action = service->a_state;

    /*D_("history_add_values() service : %s, name: %s, action: %s\n", service->name, NULL, service->a_state->state_name); */

    add_hist(tmp_e);

    /* if service is status freeing, clear the pointers in history db */
    if (IS_MARK(service, &FREEING))
    {
        history_db_clear_service(service);
    }

    /* leave */
    return (TRUE);
}

static void fetch_output(active_h * service, int datalen, process_h * process)
{
    history_h *tmp_e = NULL;

    /* allocate space for data */
    if (!(tmp_e = (history_h *) initng_calloc(1, sizeof(history_h))))
    {
        F_("Out of memory.\n");
        return;
    }

    /* check data lenght */
    if (datalen < 1)
    {
        free(tmp_e);
        return;
    }

    /* set data in struct */
    tmp_e->service = service;
    tmp_e->name = NULL;
    tmp_e->time = time(0);
    tmp_e->data = i_strdup(&process->buffer[process->buffer_pos]);
    tmp_e->action = NULL;

    /* add to history struct */
    add_hist(tmp_e);
}

static void cmd_log(char *arg, FILE * fd)
{
    char *name = NULL;
    char *latest = NULL;
    struct tm *ts;
    int only_output = FALSE;
    history_h *current = NULL;

    /* if arg is "output" only print service output */
    if (strcmp(arg, "output") == 0)
        only_output = TRUE;

    /* reset arg, if strlen is short */
    if (strlen(arg) < 1)
        arg = NULL;

    fprintf(fd, " hh:mm:ss    service           :  STATUS\n");
    fprintf(fd, " ------------------------------------------------------\n");
    while_history_db_prev(current)
    {
        /* if only_output is set, it have to be an output to continue */
        if (only_output && !current->data)
            continue;

        /* only log a few states set */
        if (!current->data && (current->action != &STARTING &&
                               current->action != &FAIL_STARTING &&
                               current->action != &RUNNING &&
                               current->action != &DONE &&
                               current->action != &STOPPING &&
                               current->action != &FAIL_STOPPING &&
                               current->action != &STOPPED))
            continue;

        /* if there is an argumet, it have to match the service */
        if (arg
            && !(((current->name) && strcmp(current->name, arg) == 0)
                 || ((current->service && current->service->name)
                     && strcmp(current->service->name, arg) == 0)))
            continue;

        /* try to set the service name, from any source found */
        if (current->name)
            name = current->name;
        else if (current->service && current->service->name)
            name = current->service->name;
        else
            name = NULL;

        /* create an localtime struct form service->time */
        ts = localtime(&current->time);

        /* if the log entry contains output data ... */
        if (current->data)
        {
            if (latest == name)
                fprintf(fd, "%s", current->data);
            else
                fprintf(fd, " %.2i:%.2i:%.2i %20s : *OUTPUT*\n%s",
                        ts->tm_hour, ts->tm_min, ts->tm_sec, name,
                        current->data);
            latest = name;
        }
        else
        {
            /* reset so that *OUTPUT* will be shown again */
            latest = NULL;

            /* print a nice service change status entry */
            fprintf(fd, " %.2i:%.2i:%.2i %20s : %s\n", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, name,
                    current->action->state_name);
        }

    }
    fprintf(fd, "\n\n");
}

static void cmd_history(char *arg, FILE * fd)
{
    history_h *current = NULL;
    active_row row;

    memset(&row, 0, sizeof row);

    while_history_db_prev(current)
    {
        /* if action is not set, it is probably a string logged in this db */
        if (!current->action)
            continue;

        if ((arg && strlen(arg) > 1)
            && !(((current->name) && strcmp(current->name, arg) == 0)
                 || ((current->service && current->service->name)
                     && strcmp(current->service->name, arg) == 0)))
            continue;

        if (current->data)
            printf("%s\n", current->data);

        if (current->action)
            strncpy(row.s, current->action->state_name, 100);
        else
            row.s[0] = '\0';

        row.t = current->time;
        if (current->name)
            strncpy(row.n, current->name, 100);
        else if (current->service && current->service->name)
            strncpy(row.n, current->service->name, 100);
        else
            row.n[0] = '\0';

        if (fwrite(&row, sizeof(active_row), 1, fd) != 1)
            F_("Failed to write all data\n");
    }
}


int module_init(const char *version)
{
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    INIT_LIST_HEAD(&history_db.list);

    initng_command_add(&HISTORYS);
    initng_command_add(&LOG);
    initng_add_hook(ASTATUS_CHANGE, 1000, &history_add_values);
    initng_add_hook(COMPENSATE_TIME, 50, &history_db_compensate_time);
    initng_add_hook(PIPEWATCHERS, 50, &fetch_output);

    return (TRUE);
}


void module_unload(void)
{
    initng_command_del(&HISTORYS);
    initng_command_del(&LOG);
    history_free_all();
    initng_del_hook(ASTATUS_CHANGE, &history_add_values);
    initng_del_hook(COMPENSATE_TIME, &history_db_compensate_time);
    initng_del_hook(PIPEWATCHERS, &fetch_output);
}
