
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_common.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_env_variable.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_static_process_types.h"
#include "../../src/initng_service.h"

#include "initng_pidfile.h"

/* The name of the process, initng should probe */
s_entry PIDOF = { "pid_of", STRING, 6, &TYPE_DAEMON,
    "When daemon exits, initng will look for a process with this name, and set daemon pid no to that pid."
};

/* The filename/path of a pidfile that initng can fetch pid no */
s_entry PIDFILE = { "pid_file", STRING, 8, &TYPE_DAEMON,
    "When daemon exits, initng will get pid of daemon from this file."
};

/*
 * this is an internal option, that is set after the
 * service died once, and the pid is updated.
 */
s_entry INTERNAL_GOT_PID = { NULL, SET, 0, NULL, NULL };

static pid_t pid_of(const char *name)
{
    DIR *dir;
    struct dirent *d;
    FILE *fp;

#define BUFF_SIZE 512

    D_("Will check for: %s\n", name);

    dir = opendir("/proc");
    if (!dir)
        return (-1);

    /* Walk through the directory. */
    while ((d = readdir(dir)) != NULL)
    {
        char buf[BUFF_SIZE + 1];
        char *s = NULL;
        int len = 0;
        pid_t pid = -1;

        /* See if this is a process */
        if ((pid = atoi(d->d_name)) <= 0)
            continue;

        /* Open the status file. */
        snprintf(buf, BUFF_SIZE, "/proc/%d/stat", pid);
        D_("To open: %s\n", buf);

        /* Read SID & statname from it. */
        if (!(fp = fopen(buf, "r")))
        {
            W_("Could not open %s.\n", buf);
            continue;
        }

        /* HACK: ignore errors */
        if (fgets(buf, BUFF_SIZE, fp) == NULL)
            buf[0] = '\0';

        s = buf;

        /* skip to the first space */
        while (*s && *s != ' ')
            s++;
        if (*s == '\0')
            continue;

        /* skip the space */
        s++;

        /* skipp the '(' char */
        if (*s != '(')
            continue;
        s++;

        /* count the lenght */
        while (s[len] && s[len] != ')')
            len++;

        if (strncmp(s, name, len) == 0)
        {
            D_("Found %s with pid %d\n", name, pid);
            return (pid);
        }
    }

    D_("Did not find a process with name %s\n", name);
    return (-1);
}


/*
 * When trying to start a service, don't allow it to
 * be set RUNNING if there must exist a pidfile, and
 * pid entry has to be updated.
 */
static int check_if_done(active_h * s, a_state_h * status)
{
    assert(s);
    assert(s->name);

    S_;

    /* only check if pidfile exists, if we are trying to start something */
    if (status != &RUNNING)
        return (TRUE);

    if (active_db_is(&INTERNAL_GOT_PID, s))
        return (TRUE);

    /* Make sure the service uses a pidfile */
    if (active_db_is(&PIDFILE, s) || active_db_is(&PIDOF, s))
    {
        /* the service have to die, and check_on_kill have to collect the pid first */
        return (FALSE);
    }
    return (TRUE);
}

/*
 * Check if a pidfile exists, if it exists, update the 
 * pid in the active_db entry. and return TRUE
 */
static pid_t pid_from_file(const char *name)
{
    int fd = 0;
    int len = 0;
    char buf[21];

    assert(name);

    /* open pid file */
    fd = open(name, O_RDONLY);

    /* If we cant open pidfile, this is bad */
    if (fd == -1)
    {
        W_("Unable to open pidfile: %s, \"%s\", it might not be created yet.\n", name, strerror(errno));
        return (-1);
    }

    /* Read data from buffer */
    len = read(fd, buf, 20);
    close(fd);
    if (len < 1)
    {
        F_("Read 0 chars from %s, It's empty.\n", name);
        return (-1);
    }

    /* remove last newline */
    if (buf[len - 1] == '\n')
        buf[len - 1] = 0;
    else
        buf[len] = 0;

    /* Try to convert pid to int */
    return (atoi(buf));
}

static pid_t get_pidof(active_h * s)
{
    pid_t pid;
    const char *pidof;
    char *pidof_fixed = NULL;

    pidof = active_db_get_string(&PIDOF, s);
    if (!pidof)
        return (-1);


    pidof_fixed = fix_variables(pidof, s);
    if (!pidof_fixed)
        return (-1);

    pid = pid_of(pidof_fixed);
    free(pidof_fixed);
    return (pid);
}

/* this will get the pid of PIDFILE entry of service */
static pid_t get_pidfile(active_h * s)
{
    pid_t pid;
    const char *pidfile = NULL;
    char *pidfile_fixed = NULL;

    /* get the pidfile */
    pidfile = active_db_get_string(&PIDFILE, s);

    /* check so that its found */
    if (!pidfile)
        return (-1);

    /* fix the variables in the string */
    pidfile_fixed = fix_variables(pidfile, s);

    /* check so we got the string */
    if (!pidfile_fixed)
        return (-1);

    /* get the pid from the file */
    pid = pid_from_file(pidfile_fixed);
    free(pidfile_fixed);

    /* return the pid */
    return (pid);
}

static int check_on_kill(active_h * s, process_h * p)
{
    assert(s);
    assert(s->name);
    assert(p);

    pid_t pid = -1;

    S_;

    /* make sure its the start process this is about */
    if (p->pt != &T_DAEMON && s->type != &TYPE_DAEMON)
        return (FALSE);

    /* make sure this service is launching, else we dont do anythinh */
    if (!IS_MARK(s, &STARTING))
        return (FALSE);

    D_("will handle %s status %s\n", s->name, s->a_state->state_name);
    /* if rcode indicate failure, dont check for pidfile */
    if (p->r_code > 0)
        return (FALSE);
    /* check if string PIDOF or PIDFILE exits */
    if (active_db_is(&PIDOF, s))
    {
        D_("getting pid by PIDOF!\n");
        /* get pid by process name */
        pid = get_pidof(s);
        D_("result : %d\n", pid);
    }

    if (active_db_is(&PIDFILE, s))
    {
        D_("getting pid by PIDFILE!\n");
        pid = get_pidfile(s);
        D_("result : %d\n", pid);
    }

    if (pid < 2)
    {
        return (FALSE);
    }

    /* set the new pid, to that we got from pidof or pidfile */
    p->pid = pid;
    /* this will alter check_if_done abowe */
    active_db_set(&INTERNAL_GOT_PID, s);
    mark_service(s, &RUNNING);
    return (TRUE);
}

int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&PIDFILE);
    initng_sdt_add(&PIDOF);
    initng_sdt_add(&INTERNAL_GOT_PID);
    initng_add_hook(CONFIRM_ASTATUS_CHANGE, 90, &check_if_done);
    initng_add_hook(HANDLE_KILLED, 10, &check_on_kill);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&PIDFILE);
    initng_sdt_del(&PIDOF);
    initng_sdt_del(&INTERNAL_GOT_PID);
    initng_del_hook(CONFIRM_ASTATUS_CHANGE, &check_if_done);
    initng_del_hook(HANDLE_KILLED, &check_on_kill);
}
