
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include <assert.h>
#include <fnmatch.h>
#include <time.h>

#include "initng_global.h"
#include "initng_active_db.h"
#include "initng_process_db.h"
#include "initng_service_cache.h"
#include "initng_toolbox.h"
#include "initng_common.h"
#include "initng_static_data_id.h"
#include "initng_plugin_callers.h"
#include "initng_string_tools.h"
#include "initng_static_states.h"

/* #####  ACTIVE_DB_FIND_BY   ######## */

/*
 * This is an exact search "net/eth0" must be "net/eth0"
 */
active_h *active_db_find_by_exact_name(const char *service)
{
    active_h *current = NULL;

    D_("(%s);\n", (char *) service);

    assert(service);

    /* first, search for an exact match */
    while_active_db(current)
    {
        assert(current->name);
        /* check if this service name is like service */
        if (strcmp(current->name, service) == 0)
            return (current);               /* return it */
    }

    /* did not find any */
    return NULL;
}

/*
 * This is an less exact name, that will work with wildcards.
 * Searching for "net/eth*" will give you "net/eth0"
 */
active_h *active_db_find_by_name(const char *service)
{
    assert(service);
    active_h *current = NULL;

    D_("(%s);\n", (char *) service);

    /* first give the exact find a shot */
    if ((current = active_db_find_by_exact_name(service)))
        return (current);

    /* walk the active db and compere */
    current = NULL;
    while_active_db(current)
    {
        assert(current->name);
        /* then try to find alike name */
        if (service_match(current->name, service))
            return (current);

    }

    /* did not find any */
    return NULL;
}

/*
 * This will search and find the best possible.
 * Search for "eth" will get you "net/eth0"
 */
active_h *active_db_find_in_name(const char *service)
{
    active_h *current = NULL;
    char *copy = NULL;
    int i = 0;

    assert(service);

    D_("(%s);\n", (char *) service);

    /* first search by name */
    if ((current = active_db_find_by_name(service)))
        return (current);

    /* remove starting wildcards */
    while (service[0] == '*' || service[0] == '?')
        service++;

    copy = i_strdup(service);
    assert(copy);

    /* remove last wildcards */
    for (i = 0; copy[i]; i++)
    {
        if (copy[i] == '*' || copy[i] == '?')
            copy[i] = '\0';
    }

    /* then search for a word match */
    current = NULL;
    while_active_db(current)
    {
        assert(current->name);
        /* check if this service name is like service */
        if (strstr(current->name, copy))
        {
            D_("Warning, the return is not an exact match!\n");
            free(copy);
            return (current);               /* return it */
        }
    }
    free(copy);

    /* did not find any */
    return NULL;
}

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_service_h(service_h * service)
{
    active_h *current = NULL;

    assert(service);
    assert(service->name);

    /* walk the active_db */
    while_active_db(current)
    {
        assert(current->name);
        /* check if this service->from_service is like service */
        if (current->from_service && current->from_service == service)
            return (current);               /* return it */

    }

    return NULL;
}

/* returns pointer to active_h process belongs to, and sets process_type */
active_h *active_db_find_by_pid(pid_t pid)
{
    active_h *currentA = NULL;
    process_h *currentP = NULL;

    /* walk the active_db */
    while_active_db(currentA)
    {
        assert(currentA->name);
        currentP = NULL;
        while_processes(currentP, currentA)
        {
            if (currentP->pid == pid)
                return (currentA);
        }
    }

    /* bad luck */
    return (NULL);
}

/* #########   STRUCT/CHAIN HANDLE  ######## */

/* add active to data structure */
/* remember to free service, if this fails */
int active_db_add(active_h * add_this)
{
    active_h *current = NULL;

    /* we have to get this data */
    assert(add_this);
    assert(add_this->name);

    /* Look for duplicate */
    if ((current = active_db_find_by_name(add_this->name)))
    {
        /* TODO, should add_this bee freed? */
        W_("active_db_add(%s): duplicate here\n", add_this->name);
        return (FALSE);
    }

    list_add(&add_this->list, &g.active_database.list);

    return (TRUE);
}

/* ############    ACTIVE_DB_ALLOC/FREE ##########  */

/* creates a new active_h at new_active */
active_h *active_db_new(const char *name)
{
    active_h *new_active = NULL;

    assert(name);

    /* allocate a new active entry */
    new_active = (active_h *) initng_calloc(1, sizeof(active_h));       /* Allocate memory for a new active */

    /* out of memory? */
    if (!new_active)
    {
        F_("Unable to allocate active, out of memory?\n");
        return (NULL);
    }

    new_active->name = i_strdup(name);
    if (!new_active->name)
    {
        F_("Unable to set name, out of memory?\n");
        return (NULL);
    }

    INIT_LIST_HEAD(&(new_active->data.list));
    INIT_LIST_HEAD(&(new_active->processes.list));

    new_active->time_got_status = time(0);
    new_active->uptime = new_active->time_got_status;
    new_active->a_state = NULL;             /* a_state should be NULL to mark an emtpy service */
    new_active->from_service = NULL;


    return (new_active);
}

/* free some values in this one */
void active_db_free(active_h * pf)
{
    process_h *current = NULL;
    process_h *safe = NULL;

    assert(pf);
    assert(pf->name);

    D_("(%s);\n", pf->name);

    /* look if there are plugins, that is intrested to now, this is freeing */
    mark_service(pf, &FREEING);

    while_processes_safe(current, safe, pf)
    {
        list_del(&current->list);
        process_db_free(current);
    }

    /* remove every data entry */
    d_remove_all(&pf->data.list);

    /* remove file cache of entry if present, so we got a fresh read from file when this service is restarted */
    if (pf->from_service)
    {
        /* remove from cache list */
        list_del(&pf->from_service->list);
        /* free entry */
        service_db_free(pf->from_service);
    }

    /* free service name */
    if (pf->name)
        free(pf->name);

    /* free service struct */
    free(pf);
}


/* clear database */
void active_db_free_all(void)
{
    active_h *current, *safe = NULL;

    while_active_db_safe(current, safe)
    {
        assert(current->name);
        list_del(&current->list);
        active_db_free(current);
    }
}

/* #########  ACTIVE_DB_UTILS  #############  */

/*
 * active_db_dep_on:
 *  Will check with all plugins and return TRUE
 *  if service depends on check.
 */
int active_db_dep_on(active_h * service, active_h * check)
{
    s_call *current, *s = NULL;
    int result = FALSE;

    assert(service);
    assert(check);

    if (service == check)
        return (FALSE);

    while_call_db_safe(current, s)
    {
        if (current->ptype != DEP_ON)
            continue;

        if ((result = (*current->c.dep_on_check) (service, check) == TRUE))
            break;
    }
    return result;
}




/*
 * A deeper deepfind.
 * Logic, we wanna make sure that depend check in a deeper level.
 * if daemon/smbd -> daemon/samba -> system/checkroot -> system/initial.
 * So should active_db_dep_on_deep(daemon/smbd, system/initial) == TRUE
 *
 * Sumary, does service depends on check?
 */
static int active_db_dep_on_deep2(active_h * service, active_h * check,
                                  int *stack);
int active_db_dep_on_deep(active_h * service, active_h * check)
{
    int stack = 0;

    return (active_db_dep_on_deep2(service, check, &stack));
}

static int active_db_dep_on_deep2(active_h * service, active_h * check,
                                  int *stack)
{
    active_h *current = NULL;
    int result = FALSE;

    if (current == service)
        return (FALSE);

    /* if service depends on check, it also dep_on_deep's on check */
    /* this serves as an exit from the recusrion */
    if (active_db_dep_on(service, check))
        return (TRUE);

    /* in case there is a circular dependency, break after 10 levels of recusrion */
    (*stack)++;
    if (*stack > 10)
        return (FALSE);

    /* loop over all services, if service depends on current, recursivly check if
     * current may depend (deep) on check */
    while_active_db(current)
    {
        if (active_db_dep_on(service, current))
        {
            if ((result = active_db_dep_on_deep(current, check)))
                break;
        }
    }
    return result;
}

/* compensate time */
void active_db_compensate_time(time_t skew)
{
    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        assert(current->name);
        /* change this time */
        current->time_got_status += skew;
        current->uptime += skew;
    }
}


/* update all service_h pointers in whole active_db */
void active_db_change_service_h(service_h * from, service_h * to)
{
    active_h *current = NULL;

    assert(from);
    /* !assert(to) to can be NULL, if we wana clear that entry */

    /* walk the active_db */
    while_active_db(current)
    {
        assert(current->name);
        /* change this time */
        if (current->from_service == from)
            current->from_service = to;
    }
}

/* active_db_count counts a type, if null, count all */
int active_db_count(a_state_h * a_state_to_count)
{
    int counter = 0;            /* actives counter */
    active_h *current = NULL;

    /* ok, go COUNT ALL */
    if (!a_state_to_count)
    {
        /* ok, go thru all */
        while_active_db(current)
        {
            assert(current->name);

            /* count almost all */

            /* but not failed services */
            if (IS_FAILED(current))
                continue;
            /* and not stopped */
            if (IS_DOWN(current))
                continue;

            counter++;
        }

        return (counter);
    }

    /* ok, go COUNT A SPECIAL */
    while_active_db(current)
    {
        assert(current->name);
        /* check if this is the status to count */
        if (current->a_state == a_state_to_count)
            counter++;
    }
    /* return counter */
    return (counter);
}


/* calculate percent of processes started */
int active_db_percent_started(void)
{
    int starting = 0;
    int up = 0;
    int other = 0;
    float tmp = 0;

    active_h *current = NULL;

    /* walk the active_db */
    while_active_db(current)
    {
        assert(current->name);

        /* if a_state is not set yet, it is probably werry starting */
        if (!current->a_state)
        {
            starting++;
            continue;
        }

        /* count starting */
        if (IS_STARTING(current))
        {
            starting++;
            continue;
        }

        /* count up */
        if (IS_UP(current))
        {
            up++;
            continue;
        }

        /* count others */
        other++;
    }
    D_("active_db_percent_started(): up: %i   starting: %i  other: %i\n", up,
       starting, other);

    /* if no one starting */
    if (starting <= 0)
        return (100);

    if (up > 0)
    {
        tmp = 100 * (float) up / (float) (starting + up);
        D_("active_db_percent_started(): up/starting: %f percent: %i\n\n",
           (float) up / (float) starting, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}

/* calculate percent of processes stopped */
int active_db_percent_stopped(void)
{
    int stopping = 0;
    int down = 0;
    int other = 0;
    float tmp = 0;
    active_h *current = NULL;

    while_active_db(current)
    {
        assert(current->name);

        /* if it has no a_status it is probable loading, or something bad happend */
        if (!current->a_state)
        {
            other++;
            continue;
        }

        /* count stopped services */
        if (IS_DOWN(current))
        {
            down++;
            continue;
        }

        /* count stopping */
        if (IS_STOPPING(current))
        {
            stopping++;
            continue;
        }

        /* count others */
        other++;
    }

    D_("active_db_percent_stopped(): down: %i   stopping: %i  other: %i\n",
       down, stopping, other);

    /* if no one stopping */
    if (stopping <= 0)
        return (100);

    if (down > 0)
    {
        tmp = 100 * (float) down / (float) (stopping + down);
        D_("active_db_percent_stopped(): down/stopping: %f percent: %i\n\n",
           (float) down / (float) stopping, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}


/* ########     Data handlers  ############## */
const char *active_db_get_string_var(s_entry * type, const char *vn,
                                     active_h * from_active)
{
    const char *t = NULL;

    assert(from_active);

    /* look in active datas */
    if ((t = d_get_string_var(type, vn, &from_active->data.list)))
        return (t);

    /* load a service file, if not exists */
    if (!from_active->from_service)
        get_service(from_active);

    /* look in service file data */
    if (from_active->from_service)
        return (service_db_get_string_var
                (type, vn, from_active->from_service));

    return (NULL);
}

int active_db_get_int_var(s_entry * type, const char *vn,
                          active_h * from_active)
{
    assert(from_active);

    if (d_is_var(type, vn, &from_active->data.list))
        return (d_get_int_var(type, vn, &from_active->data.list));

    if (!from_active->from_service)
        get_service(from_active);
    if (from_active->from_service)
        return (service_db_get_int_var
                (type, vn, (from_active->from_service)));
    return (0);
}

const char *active_db_get_next_string_var(s_entry * type, const char *vn,
                                          active_h * from_active,
                                          const char *last)
{
    const char *r = NULL;

    assert(from_active);

    /* check active overlay data */
    if ((r = d_get_next_string_var(type, vn, &from_active->data.list, last)))
        return (r);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        return (service_db_get_next_string_var
                (type, vn, from_active->from_service, last));
    return (NULL);
}


int active_db_is_var(s_entry * type, const char *vn, active_h * from_active)
{
    assert(from_active);

    /* check active overlay data */
    if (d_is_var(type, vn, &from_active->data.list))
        return (TRUE);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        if (service_db_is_var(type, vn, from_active->from_service))
            return (TRUE);

    return (FALSE);
}

int active_db_count_type(s_entry * type, active_h * from_active)
{
    assert(from_active);
    if (!from_active->from_service)
        get_service(from_active);

    if (from_active->from_service)
        return (service_db_count_type(type, from_active->from_service));
    return (0);
}
