/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>
#include <string.h>
#include <dirent.h>
#include <fnmatch.h>
#include <errno.h>
#include "initng.h"
#include "initng_global.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_module.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_common.h"
#include "../initng-paths.h"

/* 
 * Some modules depend on other modules.
 *
 * A module declares its dependencies by adding an array called
 * module_needs, with the names of the modules that it needs, like
 * this:
 *
 * char *module_needs[] = {
 *     "foo",
 *     "bar",
 *     NULL
 * };
 *
 * When a module is loaded, it checks to see if all of the modules
 * that it depends on are already loaded. If they are not, it refuses
 * to load.
 *
 * The initng_load_all_modules() routine will insure that modules are
 * loaded in the correct order. It does this by loading all modules
 * that have their dependencies met, and iterating until either all
 * modules are loaded or there are no modules left to load. (This is
 * effectively a breadth-first search.)
 *
 * Unloading follows the same logic, except in reverse. A module may
 * not be unloaded if a module that depends on it is loaded. Likewise,
 * the initng_unload_all_modules() routine will unload modules in the
 * proper order. (Again, through a breadth-first search.)
 *
 * Modules that result in circular dependencies can never be loaded.
 *
 * Several things are missing from the depencency setup:
 *
 * - Some notion of "services". A module can depend on only specific
 *   module, say "cpout", not "any module that provides output".
 *
 * - Priority mechanism. The callback functions from modules are
 *   invoked in the order that they are loaded.
 *
 * - Ability for modules to refuse to unload.
 * 
 * Hopefully these will not matter!
 */

/* Note: modules are added to the *start* of the linked list when they
 * are loaded. This makes unloading all modules easy, because a module
 * always sits before the modules that it depends on in the list. */

/* Static includes */
#ifdef STATIC_CHROOT
#include "../plugins/chroot/initng_chroot.h"
#endif


/*
 * See if we have already loaded a module with the given name.
 *
 * Returns TRUE if already loaded, else FALSE.
 */
static int module_is_loaded(const char *module_name)
{
    m_h *m = NULL;

    assert(module_name != NULL);

    while_module_db(m)
    {
        if (strcmp(m->module_name, module_name) == 0)
        {
            return TRUE;
        }
    }
    return FALSE;
}

/* 
 * Returns TRUE if module has its dependencies met, else FALSE.
 */
static int module_needs_are_loaded(const m_h * m)
{
    char **needs;
    int retval;

    assert(m != NULL);

    /* if there are no needs, then we have met them */
    if (m->module_needs == NULL)
    {
        return TRUE;
    }

    /* otherwise check each one */
    needs = m->module_needs;
    retval = TRUE;
    if (needs != NULL)
    {
        while (*needs != NULL)
        {
            if (!module_is_loaded(*needs))
            {
                D_("Module \"%s\" needs \"%s\"\n", m->module_name, *needs);
                retval = FALSE;
            }
            needs++;
        }
    }
    return retval;
}

/* 
 * Read the module information from the file. Does not actually call
 * module_init() for the module, so it is not "loaded" at this point.
 */
m_h *open_module(const char *module_path, const char *module_name)
{
    struct stat st;
    char *errmsg;
    m_h *m = NULL;

    assert(module_path != NULL);
    assert(module_name != NULL);

    /* allocate, the new module info struct */
    if (!(m = (m_h *) initng_calloc(1, sizeof(m_h))))
    {
        F_("Unable to allocate memory, for new module description.\n");
        return (NULL);
    }

    m->initziated = FALSE;

    /* check that file exists */
    if (stat(module_path, &st) != 0)
    {
        F_("Module \"%s\" not found\n", module_path);
        free(m);
        return (NULL);
    }

    /* open module */
    dlerror();                              /* clear any existing error */
    m->module_dlhandle = dlopen(module_path, RTLD_LAZY);
    /* 
     * this breaks ngc2 on my testbox - neuron : 
     * g.modules[i].module = dlopen(module_name, RTLD_NOW | RTLD_GLOBAL); 
     * */
    if (m->module_dlhandle == NULL)
    {
        F_("Error opening module %s; %s\n", module_name, dlerror());
        free(m);
        return (NULL);
    }

    D_("Success opening module \"%s\"\n", module_name);

    /* get intialisation function */
    dlerror();                              /* clear any existing error */
    m->module_init = dlsym(m->module_dlhandle, "module_init");
    if (m->module_init == NULL)
    {
        errmsg = dlerror();
        F_("Error reading module_init(); %s\n", errmsg);
        close_and_free_module(m);
        return (NULL);
    }

    /* get unload function */
    dlerror();                              /* clear any existing error */
    m->module_unload = dlsym(m->module_dlhandle, "module_unload");
    if (m->module_unload == NULL)
    {
        errmsg = dlerror();
        F_("Error reading module_unload(); %s\n", errmsg);
        close_and_free_module(m);
        return (NULL);
    }

    /* get dependency list (may be NULL - this is not an error) */
    dlerror();                              /* clear any existing error */
    m->module_needs = (char **) dlsym(m->module_dlhandle, "module_needs");
    /* XXX: is there any way to check for "not found", since we don't 
     * consider that an error? */

    /* set module name in database */
    m->module_name = i_strdup(module_name);

    return (m);
}

/* 
 * Close the module.
 */
void close_and_free_module(m_h * m)
{
    assert(m != NULL);

    /* free module name */
    if (m->module_name)
    {
        free(m->module_name);
        m->module_name = NULL;
    }

    /* close the lib */
    if (m->module_dlhandle)
        dlclose(m->module_dlhandle);

    /* remove from list if added */
    list_del(&m->list);

    /* free struct */
    free(m);
}

/* load a dynamic module */
/* XXX: more information! */
m_h *initng_load_module(const char *module_name)
{
    char *module_path;
    m_h *new_m;

    assert(module_name != NULL);


    /* look for duplicates */
    if (module_is_loaded(module_name))
    {
        F_("Module \"%s\" already loaded, won't load it twice\n",
           module_name);
        return (NULL);
    }


    /* build a path */
    module_path =
        (char *) initng_calloc(1,
                               strlen(INITNG_PLUGIN_DIR) +
                               strlen(module_name) + 9);
    strcpy(module_path, INITNG_PLUGIN_DIR "/lib");
    strcat(module_path, module_name);
    strcat(module_path, ".so");

    /* load information from library */
    new_m = open_module(module_path, module_name);
    free(module_path);

    if (!new_m)
    {
        F_("Unable to load module \"%s\"\n", module_name);
        return (NULL);
    }

    /* see if we have our dependencies met */
    if (!module_needs_are_loaded(new_m))
    {
        F_("Not loading module \"%s\", missing needed module(s)\n",
           module_path);
        close_and_free_module(new_m);
        return (NULL);
    }

    /* run module_init */
    new_m->initziated = (*new_m->module_init) (INITNG_VERSION);

    D_("for module \"%s\" return: %i\n", module_path, new_m->initziated);

    if (new_m->initziated < 1)
    {
        F_("Module %s did not load correctly (module_init() returned %i)\n",
           module_path, new_m->initziated);
        /* XXX: used to be here, but why? */
        /* sleep(1); */
        close_and_free_module(new_m);
        return (NULL);
    }

    assert(new_m->module_name);
    list_add(&new_m->list, &g.module_db.list);
    /* and we're done */
    return (new_m);
}



/*
 * See the named module is needed elsewhere.
 *
 * Returns TRUE if the module is needed, else FALSE.
 */
static int module_is_needed(const char *module_name)
{
    char **needs;
    m_h *m = NULL;
    int retval = FALSE;

    assert(module_name != NULL);

    while_module_db(m)
    {
        /* if not this module, have needs set, continue.. */
        if (!(m->module_needs))
            continue;
        needs = m->module_needs;

        while (*needs != NULL)
        {
            if (strcmp(module_name, *needs) == 0)
            {
                F_("Module \"%s\" needed by \"%s\"\n", module_name,
                   m->module_name);
                retval = TRUE;
            }
            needs++;
        }
    }
    return retval;
}

/*
 * Invoke callback, clean up module.
 */
static void initng_unload_module(m_h * module)
{
    assert(module != NULL);

    /* run the unload hook */
    (*module->module_unload) ();

    /* close and free the module entry in db */
    close_and_free_module(module);
}

/* XXX: more information! */
int initng_unload_module_named(const char *name)
{
    m_h *m = NULL;

    assert(name != NULL);

    D_("initng_load_module_named(%s);\n", name);




    if (!module_is_loaded(name))
    {
        F_("Not unloading module \"%s\", it is not loaded\n", name);
        return FALSE;
    }


    /* find the named module in our linked list */

    while_module_db(m)
    {
        if (strcmp(m->module_name, name) == 0)
        {
            m->marked_for_removal = TRUE;
            g.modules_to_unload = TRUE;
            return (TRUE);
        }
    }

    return FALSE;
}

int initng_load_all_modules(void)
{
    DIR *d;
    struct dirent *e;
    char *module_path;
    char *module_name;

    m_h *current, *safe = NULL;

    /* open plugin dir */
    if (!(d = opendir(INITNG_PLUGIN_DIR)))
    {
        F_("Unable to open plugin directory " INITNG_PLUGIN_DIR ".\n");
        return FALSE;
    }

    /* memory for full path */
    module_path = initng_calloc(strlen(INITNG_PLUGIN_DIR) + NAME_MAX + 2, 1);

    /* get every entry */
    while ((e = readdir(d)) != NULL)
    {
        /* check for files, ending with .so */
        if (fnmatch("lib*.so", e->d_name, 0) == 0)
        {
            module_name = i_strndup(e->d_name + 3, strlen(e->d_name + 3) - 3);

            /* search the plugin name, for blacklisted */
            if (service_blacklisted(module_name))
            {
                F_("Plugin %s blacklisted.\n", module_name);
                free(module_name);
                continue;
            }

            /* set full path */
            strcpy(module_path, INITNG_PLUGIN_DIR "/");
            strcat(module_path, e->d_name);

            /* open the module */
            current = open_module(module_path, module_name);

            /* add this to the list of loaded modules */
            if (!current)
            {
                free(module_name);
                module_name = NULL;
                continue;
            }

            /* add to list and continue */
            assert(current->module_name);
            list_add(&current->list, &g.module_db.list);
        }
        else
        {
#ifdef DEBUG
            if (e->d_name[0] != '.')
            {
                D_("Won't load module \"%s\", doesn't match \"*.so\" pattern.\n", e->d_name);
            }
#endif
        }
    }
    closedir(d);

    free(module_path);

    /* load the entries on our todo list */
    while_module_db_safe(current, safe)
    {
        /* allredy initziated */
        if (current->initziated == TRUE)
            continue;

        if (!module_needs_are_loaded(current))
        {
            close_and_free_module(current);
            continue;
        }

        /* if we did find find a module with needs loaded, try to load it */
        current->initziated = (*current->module_init) (INITNG_VERSION);
        D_("for module \"%s\" return: %i\n", current->module_name,
           current->initziated);

        /* check if it was initziated correctly */
        if (current->initziated != TRUE)
        {
            F_("Module %s did not load correctly (module_init() returned %i)\n", current->module_name, current->initziated);
            close_and_free_module(current);
        }

    }

    /*    initng_load_static_modules(); */
    return TRUE;
}


/* this will dlclose all arrays */
/* XXX: more information needed */

/* 
 * We can simply unload modules in order, because 
 */
void initng_unload_all_modules(void)
{
    m_h *m, *safe = NULL;

    while_module_db_safe(m, safe)
    {
        initng_unload_module(m);
    }

    /* reset the list, to make sure its empty */
    INIT_LIST_HEAD(&g.module_db.list);

    D_("initng_load_module_close_all()\n");
}

/* 
 * We can simply unload marked modules in order.
 */
void initng_unload_marked_modules(void)
{
    m_h *m, *safe = NULL;

    S_;



    while_module_db_safe(m, safe)
    {
        if (m->marked_for_removal == TRUE)
        {
            if (module_is_needed(m->module_name))
            {
                F_("Not unloading module \"%s\", it is needed\n",
                   m->module_name);
                m->marked_for_removal = FALSE;
                continue;
            }
            D_("now unloading marked module %s.\n", m->module_name);

            initng_unload_module(m);
        }
    }
}
